import json
import numpy as np
import pandas as pd
import altair as alt
import scipy.stats as stats
from typing import Optional
from mosqlient import get_prediction_by_id
from scoringrules import crps_normal, logs_normal
from sklearn.metrics import mean_squared_error, mean_absolute_error

def transform_json_to_dataframe(res: dict) -> pd.DataFrame:
    """
    A function that transforms the prediction output from the API and
    transforms it in a DataFrame.

    Parameters:
    rest (dict): Output of the  prediction's API.

    Returns:
    pd.DataFrame.
    """

    json_struct = json.loads(res['prediction'])
    df = pd.json_normalize(json_struct)
    df.dates = pd.to_datetime(df.dates)

    return df


def evaluate_point_metrics(y_true, y_pred, metric):
    """
    Evaluate multiple sklearn metrics on given true and predicted values.

    Parameters:
    -------------
    y_true (array-like): True values.
    y_pred (array-like): Predicted values.
    metrics (str): Options: ['MAE', 'MSE'] .

    Returns:
    Scores.
    """

    if metric == 'MAE':

        m = mean_absolute_error

    if metric == 'MSE':

        m = mean_squared_error

    score = m(y_true, y_pred)

    return score


def compute_interval_score(lower_bound, upper_bound, observed_value, alpha=0.05):
    """
    Calculate the interval score for a given prediction interval and observed value.

    Parameters:
    ------------------
    lower_bound: float | np.array 
        The lower bound of the prediction interval.
    upper_bound: float | np.array
        The upper bound of the prediction interval.
    observed_value: float | np.array
        The observed value.
    alpha: float
        The significance level of the interval. Default is 0.05 (for 95% prediction intervals).

    Returns:
    -----------
    float or np.array: The interval score.
    """


    interval_width = upper_bound - lower_bound
    
    # Compute penalties
    penalty_lower = 2 / alpha * np.maximum(0, lower_bound - observed_value)
    penalty_upper = 2 / alpha * np.maximum(0, observed_value - upper_bound)
    
    penalty = penalty_lower + penalty_upper
    
    return interval_width + penalty



def plot_bar_score(data: pd.DataFrame, score: str) -> alt.Chart:
    '''
    Function to plot a bar chart based on scorer.summary dataframe

    Parameters:
    --------------
    data: pd.DataFrame
    score: str
        Valid options are: ['mae', 'mse', 'crps', 'log_score']
    '''
    data = data.reset_index()
    data['id'] = data['id'].astype(str)

    bar_chart = alt.Chart(data).mark_bar().encode(
        x=alt.X('id:N', axis=alt.Axis(labelAngle=360)).title('Model'),
        y=alt.Y(f'{score}:Q').title(score),
        color=alt.Color('id', legend=alt.Legend(title='Model'))
    ).properties(
        title=f'{score} score',
        width=400,
        height=300,
    )

    return bar_chart


def plot_score(
    data: pd.DataFrame,
    df_melted: pd.DataFrame,
    score: str = 'CRPS'
) -> alt.VConcatChart:
    '''
    Function that returns an Altair panel with the time series of cases and the
    time series of the score for each model.

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame with the time series of cases must contain the columns
        `dates` and `casos`.
    df_melted : pd.DataFrame
        The DataFrame must contains the columns:
        * dates: with the dates';
            * variable: with the models name;
        * '{score}_score': with the score value
    '''

    if score == 'CRPS':
        title = 'CRPS score'
        subtitle = 'Lower is better'

    if score == 'interval':
        title = 'Interval score'
        subtitle = 'Lower is better'

    if score == 'log':
        title = 'Log score'
        subtitle = 'Bigger is better'

    timedata = alt.Chart(data).mark_line().encode(
        x='dates',
        y='casos',
        color=alt.value('black')).properties(
        width=400,  # Set the width
        height=300  # Set the height
    )

    # Create a selection that chooses the nearest point & selects based on x-value
    nearest = alt.selection_point(
        nearest=True,
        on="pointerover",
        fields=["dates"],
        empty=False
    )

    graph_score = alt.Chart(df_melted).mark_point(filled=False).encode(
        x='dates',
        y=f'{score}_score',
        color=alt.Color('variable', legend=alt.Legend(legendX=100)),
    ).properties(
        width=400,  # Set the width
        height=250  # Set the height
    )

    # Transparent selectors across the chart. This is what tells us
    # the x-value of the cursor
    selectors = alt.Chart(df_melted).mark_point().encode(  # TODO: Not used
        x="dates",
        opacity=alt.value(0),
    ).add_params(
        nearest
    )

    # Draw points on the line, and highlight based on selection
    points = graph_score.mark_point().encode(
        opacity=alt.condition(nearest, alt.value(1), alt.value(0))
    )

    # Draw a rule at the location of the selection
    columns = list(df_melted.variable.unique())
    tooltip = [alt.Tooltip(c, type="quantitative", format=".2f")
               for c in columns]
    tooltip.insert(0, alt.Tooltip('dates:T', title='Date'))
    rules = alt.Chart(df_melted).transform_pivot(
        "variable",
        value=f"{score}_score",
        groupby=["dates"]
    ).mark_rule(color="gray").encode(
        x="dates",
        opacity=alt.condition(nearest, alt.value(0.3), alt.value(0)),
        tooltip=tooltip,
    ).add_params(nearest)

    return timedata.properties(
        width=400,  # Set the width
        height=150,  # Set the height
        title='New cases'
    ) & alt.layer(
        graph_score,  points, rules
    ).properties(
        title={
            'text': title,
            'subtitle': subtitle
        }
    )


class Scorer:
    """
    A class to compare the score of the models.

    Attributes
    ----------

    df_true: pd.DataFrame
        DataFrame of the cases provided by the user.


    filtered_df_true: pd.DataFrame
        DataFrame of the cases provided by the user filtered according
        to the interval of the predictions or with the `set_date_range` method .

    ids: Optional[list[int]]
        The list of the predictions id that will be compared


    dict_df_ids: dict[pd.DataFrame]
        A dict of DataFrames of the predictions. If the key is int it refers
        to the ids passed in the init. If it is `preds` it refers to the
        dataframe of the predictions provided by the user.

    filtered_dict_df_ids: dict[pd.DataFrame]
        A dict of DataFrames of the predictions. If the key is int it refers to
        the ids passed in the init. If it is `preds` it refers to the dataframe
        of the predictions provided by the user. The DataFrames are filtered
        according to the interval of the predictions or with the
        `set_date_range` method.

    min_date: str
        Min date that will include the information of the df_true and predictions.

    min_date: str
        Max date that will include the information of the df_true and predictions.

    mae : dict
        Dict where the keys are the id of the models or `preds` when a
        dataframe of predictions is provided by the user, and the values of
        the dict are the mean absolute error.

    mse: dict
        Dict where the keys are the id of the models or `preds` when a
        dataframe of predictions is provided by the user, and the values of the
        dict are the mean squared error.

    crps: tuple of dicts
        Dict where the keys are the id of the models or `preds` when a
        dataframe of predictions is provided by the user, and the values of the
        dict are the scores computed.

        The first dict contains the CRPS score computed for every predicted
        point, and the second one contains the mean values of the CRPS score
        for all the points.

        The CRPS computed assumes a normal distribution.

    log_score: tuple of dicts
        Dict where the keys are the id of the models or `preds` when a
        dataframe of predictions is provided by the user, and the values of the
        dict are the scores computed.

        The first dict contains the log score computed for every predicted
        point, and the second one contains the mean values of the log score for
        all the points.

        The log score computed assumes a normal distribution.


    interval_score: tuple of dicts
        Dict where the keys are the id of the models or `preds` when a
        dataframe of predictions is provided by the user, and the values of the
        dict are the scores computed.

        The first dict contains the interval score computed for every predicted
        point, and the second one contains the mean values of the interval score
        for all the points.

    summary: pd.DataFrame
        DataFrame where the keys are the id of the models or `preds` when a
        dataframe of predictions is provided by the user, and the columns are
        the scores: mae, mse, and the mean of crps, log_score, and interval score.


    Methods
    -------
    start_date_range():
        Train the model.
    plot_predictions():
        Function that returns an Altair panel (alt.Chart) with the time series
        of cases and the predictions for each model.
    plot_crps():
        alt.Chart: Method that returns an Altair panel with the time series of
        cases and the time series of the CRPS score for each model.
    plot_log_score():
        alt.Chart: Method that returns an Altair panel with the time series of
        cases and the time series of the log score for each model.
    plot_interval_score():
        alt.Chart: Method that returns an Altair panel with the time series of
        cases and the time series of the interval score for each model.
    plot_mae():
        alt.Chart : Bar chart of the MAE score for each prediction.
    plot_mse():
        alt.Chart : Bar chart of the MSE score for each prediction.
    """

    def __init__(
        self,
        df_true: pd.DataFrame,
        ids: Optional[list[int] | list[str]] = None,
        preds: Optional[pd.DataFrame] = None,
        confidence_level: float=0.90
    ):
        """
        Parameters
        ----------
        df_true: pd.DataFrame
            DataFrame with the columns `dates` and `casos`.
        ids : list[int]
            List of the predictions ids that it will be compared.
        preds: pd.DataFrame
            Pandas Dataframe already in the format accepted by the platform
            that will be computed the score.
        confidence_level: float. 
            The confidence level of the predictions of the columns upper and lower.
        """

        # input validation data
        cols_df_true = ["dates", "casos"]

        if not set(cols_df_true).issubset(set(list(df_true.columns))):
            raise ValueError(
                "Missing required keys in the df_true:"
                f"{set(cols_df_true).difference(set(list(df_true.columns)))}"
            )

        df_true.dates = pd.to_datetime(df_true.dates)
        # Ensure all the dates has the same lenght
        min_dates = [min(df_true.dates)]
        max_dates = [max(df_true.dates)]

        dict_df_ids = {}

        if preds is not None:
            cols_preds = ['dates', 'lower', 'preds', 'upper']
            if not set(cols_preds).issubset(set(list(preds.columns))):
                raise ValueError(
                    "Missing required keys in the preds:"
                    f"{set(cols_preds).difference(set(list(preds.columns)))}")

            dict_df_ids['preds'] = preds
            preds.dates = pd.to_datetime(preds.dates)
            min_dates.append(min(preds.dates))
            max_dates.append(max(preds.dates))

        if (ids is None or len(ids) == 0) and (preds is None):
            raise ValueError(
                "It must be provide and id or DataFrame to be compared"
            )

        if ids is not None:
            ids = [str(id_) for id_ in ids]
            for id_ in ids:
                prediction = get_prediction_by_id(id=int(id_))

                if not prediction:
                    raise ValueError(f"No Prediction found for id: {id_}")

                df_ = prediction.to_dataframe()
                df_ = df_.sort_values(by='dates')
                df_.dates = pd.to_datetime(df_.dates)
                dict_df_ids[id_] = df_
                min_dates.append(min(df_.dates))
                max_dates.append(max(df_.dates))

        min_dates = pd.to_datetime(min_dates)
        max_dates = pd.to_datetime(max_dates)
        min_date = max(min_dates)
        max_date = min(max_dates)

        # updating the dates interval
        df_true = df_true.loc[(df_true.dates >= min_date)
                              & (df_true.dates <= max_date)]
        df_true = df_true.sort_values(by='dates')
        df_true.reset_index(drop=True, inplace=True)

        for id_ in dict_df_ids.keys():
            df_id = dict_df_ids[id_]
            df_id = df_id.loc[(df_id.dates >= min_date) &
                              (df_id.dates <= max_date)]
            df_id = df_id.sort_values(by='dates')
            dict_df_ids[id_] = df_id

        z_value = stats.norm.ppf((1 + confidence_level) / 2)

        self.df_true = df_true
        self.filtered_df_true = df_true
        self.ids = ids
        self.dict_df_ids = dict_df_ids
        self.filtered_dict_df_ids = dict_df_ids
        self.min_date = min_date
        self.max_date = max_date
        self.confidence_level = confidence_level
        self.z_value = z_value

    def set_date_range(self, start_date: str, end_date: str) -> None:
        '''
         This method will redefine the interval of dates used to compute the
         scores.
         The new dates provided must be in the interval defined by the
         `__init__` method that ensures the df_true and predictions are in the
         same interval. You can access these values by score.min_date and
         score.max_date.

        Parameters
        --------------
        start_date: str
            The new start date used to compute the scores.
        end_date: str
            The new end date used to compute the scores.
        '''

        if (
            (self.min_date > pd.to_datetime(start_date)) or
            (self.max_date < pd.to_datetime(start_date))
        ):
            raise ValueError(
                "The start and end date must be between " +
                f"{self.min_date} and {self.max_date}."
            )

        df_true = self.df_true
        dict_df_ids = self.dict_df_ids

        self.filtered_df_true = df_true.loc[(df_true.dates >= pd.to_datetime(
            start_date)) & (df_true.dates <= pd.to_datetime(end_date))]

        for id_ in dict_df_ids.keys():
            df_id = dict_df_ids[id_]
            df_id = df_id.loc[(df_id.dates >= pd.to_datetime(start_date)) & (
                df_id.dates <= pd.to_datetime(end_date))]
            dict_df_ids[id_] = df_id

        self.filtered_dict_df_ids = dict_df_ids

        return None

    @property
    def mae(self,):
        '''
        dict: Dict, where the keys are the id of the models or `preds` when a
        dataframe of predictions is provided by the user, and the values of the
        dict are the mean absolute error.
        '''
        ids = self.ids
        dict_df_ids = self.filtered_dict_df_ids
        df_true = self.filtered_df_true

        scores = {}

        for id_ in dict_df_ids.keys():

            scores[id_] = evaluate_point_metrics(
                df_true.casos, y_pred=dict_df_ids[id_].preds, metric='MAE'
            )

        return scores

    @property
    def mse(self,):
        '''
        dict: Dict, where the keys are the id of the models or `preds` when a
        dataframe of predictions is provided by the user, and the values of the
        dict are the mean squared error.
        '''

        ids = self.ids
        dict_df_ids = self.filtered_dict_df_ids
        df_true = self.filtered_df_true

        scores = {}

        for id_ in dict_df_ids.keys():

            scores[id_] = evaluate_point_metrics(
                df_true.casos, y_pred=dict_df_ids[id_].preds, metric='MSE'
            )
        return scores

    @property
    def crps(self,):
        '''
        tuple of dict: Dict where the keys are the id of the models or `preds`
        when a dataframe of predictions is provided by the user,
        and the values of the dict are the scores computed.

        The first dict contains the CRPS score computed for every predicted
        point, and the second one contains the mean values of the CRPS score
        for all the points.

        The CRPS computed assumes a normal distribution.
        '''

        ids = self.ids
        dict_df_ids = self.filtered_dict_df_ids
        df_true = self.filtered_df_true

        scores_curve = {}

        scores_mean = {}

        for id_ in dict_df_ids.keys():

            df_id_ = dict_df_ids[id_]

            score = crps_normal(df_true.casos, df_id_.preds,
                                (df_id_.upper-df_id_.lower)/(2*self.z_value))

            scores_curve[id_] = pd.Series(score, index=df_true.dates)

            scores_mean[id_] = np.mean(score)

        self.crps_curve = scores_curve

        return scores_curve, scores_mean

    @property
    def log_score(self,):
        '''
        tuple of dict: Dict where the keys are the id of the models or `preds`
        when a dataframe of predictions is provided by the user, and the values
        of the dict are the scores computed.

        The first dict contains the log score computed for every predicted
        point, and the second one contains the mean values of the log score
        for all the points.

        The log score computed assumes a normal distribution.
        '''

        ids = self.ids
        dict_df_ids = self.filtered_dict_df_ids
        df_true = self.filtered_df_true

        scores_curve = {}
        scores_mean = {}

        for id_ in dict_df_ids.keys():

            df_id_ = dict_df_ids[id_]
            score = logs_normal(df_true.casos, df_id_.preds,
                                (df_id_.upper-df_id_.lower)/(2*self.z_value), negative=False)
            #truncated the output 
            score = np.maximum(score, np.repeat(-100, len(score)))

            scores_curve[id_] = pd.Series(score, index=df_true.dates)
            scores_mean[id_] = np.mean(score)

        self.log_curve = scores_curve

        return scores_curve, scores_mean

    @property
    def interval_score(self,):
        '''
        tuple of dict: Dict where the keys are the id of the models or `preds`
        when a dataframe of predictions is provided by the user,
        and the values of the dict are the scores computed.

        The first dict contains the interval score computed for every predicted
        point, and the second one contains the mean values of the interval score
        for all the points.
        '''

        ids = self.ids
        dict_df_ids = self.filtered_dict_df_ids
        df_true = self.filtered_df_true

        scores_curve = {}

        scores_mean = {} 

        for id_ in dict_df_ids.keys():

            df_id_ = dict_df_ids[id_]
        
            score = compute_interval_score( df_id_.lower.values,  df_id_.upper.values,df_true.casos.values,
                                alpha = 1-self.confidence_level)
            
            scores_curve[id_] = pd.Series(score, index=df_true.dates)

            scores_mean[id_] = np.mean(score)

        self.interval_score_curve = scores_curve

        return scores_curve, scores_mean
    
    @property
    def summary(self,):
        '''
        pd.DataFrame: DataFrame where the keys are the id of the models or
        `preds` when a dataframe of predictions is provided by the user, and
        the columns are the scores: mae, mse, and the mean of crps, log_score,
        and interval_score.
        '''
        sum_scores = {}

        sum_scores['mae'] = self.mae

        sum_scores['mse'] = self.mse

        sum_scores['crps'] = self.crps[1]

        sum_scores['log_score'] = self.log_score[1]

        sum_scores['interval_score'] = self.interval_score[1]

        df_score = pd.DataFrame.from_dict(sum_scores, orient='columns')

        df_score.index.name = 'id'

        return df_score

    def plot_mae(self,) -> alt.Chart:
        '''
        Bar chart of the MAE score for each prediction.
        '''

        return plot_bar_score(self.summary, 'mae')

    def plot_mse(self,) -> alt.Chart:
        '''
        Bar chart of the MSE score for each prediction.
        '''

        return plot_bar_score(self.summary, 'mse')

    def plot_crps(self,) -> alt.VConcatChart:
        '''
        alt.Chart: Function that returns an Altair panel with the time series
        of cases and the time series of the CRPS score for each model
        '''

        crps_ = self.crps_curve

        df_crps = pd.DataFrame()

        for v in crps_.keys():

            df_crps[str(v)] = crps_[v]

        df_crps.reset_index(inplace=True)

        df_melted = pd.melt(df_crps, id_vars='dates',
                            value_vars=list(map(str, crps_.keys())))
        df_melted = df_melted.rename(columns={'value': 'CRPS_score'})

        return plot_score(self.df_true, df_melted, score='CRPS')

    def plot_log_score(self,) -> alt.VConcatChart:
        '''
        alt.Chart: Function that returns an Altair panel with the time series
        of cases and the time series of the Log score for each model
        '''

        crps_ = self.log_curve

        df_crps = pd.DataFrame()

        for v in crps_.keys():

            df_crps[str(v)] = crps_[v]

        df_crps.reset_index(inplace=True)

        df_melted = pd.melt(df_crps, id_vars='dates',
                            value_vars=list(map(str, crps_.keys())))
        df_melted = df_melted.rename(columns={'value': 'log_score'})

        return plot_score(self.df_true, df_melted, score='log')

    def plot_interval_score(self,) -> alt.VConcatChart:
        '''
        alt.Chart: Function that returns an Altair panel with the time series
        of cases and the time series of the CRPS score for each model
        '''

        interval_ = self.interval_score_curve

        df_interval = pd.DataFrame()

        for v in interval_.keys():

            df_interval[str(v)] = interval_[v]

        df_interval.reset_index(inplace=True)

        df_melted = pd.melt(df_interval, id_vars='dates',
                            value_vars=list(map(str, interval_.keys())))
        df_melted = df_melted.rename(columns={'value': 'interval_score'})

        return plot_score(self.df_true, df_melted, score='interval')

    def plot_predictions(
        self,
        show_ci: bool = True,
        width: int = 400,
        height: int = 300
    ) -> alt.Chart:
        '''
        Function that returns an Altair panel (alt.Chart) with the time series
        of cases and the predictions for each model

        Parameters
        ---------------
        show_ci :bool
            If True it shows the confidence interval.
        width: int
            width of the plot
        width: int
            height of the plot
        '''

        dict_df_ids = self.filtered_dict_df_ids
        df_true_ = self.filtered_df_true
        df_true_.loc[:, 'legend'] = 'Data'

        if show_ci:
            title = "Median and 95% confidence interval"
        else:
            title = 'Median of predictions'

        df_to_plot = pd.DataFrame()

        for id_ in dict_df_ids.keys():

            df_ = dict_df_ids[id_]

            df_.loc[:, 'model'] = id_

            df_to_plot = pd.concat([df_to_plot, df_])

        df_to_plot['model'] = df_to_plot['model'].astype(str)

        data = alt.Chart(df_true_).mark_circle(size=60).encode(
            x='dates:T',
            y='casos:Q',
            color=alt.Color('legend:N', scale=alt.Scale(
                range=['black']), legend=alt.Legend(title=None))
        ).properties(
            width=width,  # Set the width
            height=height  # Set the height
        )

        # here we define the plot of the right figure
        timeseries = alt.Chart(df_to_plot, title=title).mark_line(
        ).encode(
            x=alt.X('dates:T').title('Dates'),
            y=alt.Y('preds:Q').title('New cases'),
            color=alt.Color('model', legend=alt.Legend(title='Model'))
        )

        # here we create the area that represent the confidence interval of the
        # predicitions
        timeseries_conf = timeseries.mark_area(
            opacity=0.25,
        ).encode(
            x='dates:T',
            y='lower:Q',
            y2='upper:Q',
            color=alt.Color('model', legend=None)
        )

        nearest = alt.selection_point(nearest=True, on="pointerover",
                                      fields=["dates"], empty=False)

        # Draw points on the line, and highlight based on selection
        points = timeseries.mark_point().encode(
            color=alt.Color('model', legend=None),
            opacity=alt.condition(nearest, alt.value(1), alt.value(0))
        )

        df_true_ = df_true_.rename(columns={'casos': 'preds'})

        df_true_['model'] = 'cases'

        df_to_plot = pd.concat([df_to_plot, df_true_])

        columns = list(df_to_plot.model.unique())
        tooltip = [alt.Tooltip(c, type="quantitative", format=".0f")
                   for c in columns]
        tooltip.insert(0, alt.Tooltip('dates:T', title='Date'))

        rules = alt.Chart(df_to_plot).transform_pivot(
            "model",
            value="preds",
            groupby=["dates"]
        ).mark_rule(color="gray").encode(
            x="dates",
            opacity=alt.condition(nearest, alt.value(0.3), alt.value(0)),
            tooltip=tooltip,
        ).add_params(nearest)

        if show_ci:

            final = (
                data + timeseries + timeseries_conf + points + rules
            ).resolve_scale(
                color='independent'
            )

        else:
            final = alt.layer(
                data, timeseries, points, rules
            ).resolve_scale(
                color='independent'
            )

        return final



from datetime import date
from typing import Optional, Literal

from pydantic import BaseModel

from mosqlient import types
from mosqlient.requests import get_all_sync
from mosqlient._utils import parse_params


class Infodengue(BaseModel):
    @classmethod
    def get(
        cls,
        disease: Literal["dengue", "zika", "chikungunya"],
        start: str | date,
        end: str | date,
        **kwargs
    ):
        timeout = kwargs["timeout"] if "timeout" in kwargs else 300

        kwargs["disease"] = disease
        kwargs["start"] = start
        kwargs["end"] = end
        params = parse_params(**kwargs)
        InfodengueGETParams(**params)

        return get_all_sync(
            app="datastore",
            endpoint="infodengue",
            params=params,
            pagination=True,
            timeout=timeout,
        )


class InfodengueGETParams(BaseModel):
    start: Optional[types.Date] = None
    end: Optional[types.Date] = None
    disease: Optional[types.Disease] = None
    uf: Optional[types.UF] = None
    geocode: Optional[types.Geocode] = None



UFs = {
    "AC": "Acre",
    "AL": "Alagoas",
    "AP": "Amapá",
    "AM": "Amazonas",
    "BA": "Bahia",
    "CE": "Ceará",
    "DF": "Distrito Federal",
    "ES": "Espírito Santo",
    "GO": "Goiás",
    "MA": "Maranhão",
    "MT": "Mato Grosso",
    "MS": "Mato Grosso do Sul",
    "MG": "Minas Gerais",
    "PA": "Pará",
    "PB": "Paraíba",
    "PR": "Paraná",
    "PE": "Pernambuco",
    "PI": "Piauí",
    "RJ": "Rio de Janeiro",
    "RN": "Rio Grande do Norte",
    "RS": "Rio Grande do Sul",
    "RO": "Rondônia",
    "RR": "Roraima",
    "SC": "Santa Catarina",
    "SP": "São Paulo",
    "SE": "Sergipe",
    "TO": "Tocantins",
}

UF_CODES = {
    "AC": 12,
    "AL": 27,
    "AM": 13,
    "AP": 16,
    "BA": 29,
    "CE": 23,
    "DF": 53,
    "ES": 32,
    "GO": 52,
    "MA": 21,
    "MG": 31,
    "MS": 50,
    "MT": 51,
    "PA": 15,
    "PB": 25,
    "PE": 26,
    "PI": 22,
    "PR": 41,
    "RJ": 33,
    "RN": 24,
    "RO": 11,
    "RR": 14,
    "RS": 43,
    "SC": 42,
    "SE": 28,
    "SP": 35,
    "TO": 17,
}



__all__ = ["get_infodengue"]

from typing import Optional, Literal
from datetime import date

import pandas as pd

from .infodengue import Infodengue


def get_infodengue(
    disease: Literal["dengue", "zika", "chikungunya"],
    start_date: date | str,
    end_date: date | str,
    uf: Optional[Literal[
        "AC", "AL", "AP", "AM", "BA", "CE", "ES", "GO", "MA", "MT", "MS", "MG",
        "PA", "PB", "PR", "PE", "PI", "RJ", "RN", "RS", "RO", "RR", "SC", "SP",
        "SE", "TO", "DF"
    ]] = None,
    geocode: Optional[int] = None
) -> pd.DataFrame:
    params = {
        "uf": uf,
        "geocode": geocode
    }

    return pd.DataFrame(Infodengue.get(
        disease=disease,
        start=start_date,
        end=end_date,
        **params
    ))



import json
import logging
from typing import Literal, Optional
import datetime as dt
from string import ascii_lowercase, digits

import pandas as pd

from pydantic import ValidationError

from mosqlient._config import *  # noqa
from mosqlient._utils.brasil import UF_CODES


def validate_django_app(app: str) -> str:
    assert app in DJANGO_APPS, (
        f"Unknown Mosqlimate app '{app}'. Options: {DJANGO_APPS}"
    )
    return app


def validate_id(ID: int) -> int:
    assert ID > 0, f"Incorrect ID {ID}"
    return ID


def validate_name(name: str) -> str:
    assert len(name) <= 100, "name too long"
    assert len(name) > 0, "empty name"
    return name


def validate_description(description: str) -> str:
    assert len(description) <= 500, "description too long"
    assert len(description) > 0, "empty description"
    return description


def validate_author_name(author_name: str) -> str:
    assert len(author_name) <= 100, "author_name too long"
    assert len(author_name) > 0, "empty author_name"
    return author_name


def validate_author_username(author_username: str) -> str:
    assert len(author_username) < 40, "author_username too long"
    assert len(author_username) > 0, "empty author_username"
    return author_username


def validate_author_institution(author_institution: Optional[str]) -> str | None:
    if not author_institution:
        return None
    assert len(author_institution) <= 100, "author_institution too long"
    assert len(author_institution) > 0, "empty author_institution"
    return author_institution


def validate_repository(repository: str) -> str:
    assert len(repository) <= 100, "repository too long"
    assert len(repository) > 0, "empty repository"
    return repository


def validate_implementation_language(implementation_language: str) -> str:
    languages = [
        "zig",
        "rust",
        "ruby",
        "r",
        "lua",
        "kotlin",
        "java",
        "javascript",
        "haskell",
        "go",
        "erlang",
        ".net",
        "c",
        "c#",
        "coffeescript",
        "c++",
        "python",
    ]
    assert implementation_language.lower() in languages, (
        f"Unknown implementation_language {implementation_language}"
    )
    return implementation_language


def validate_disease(disease: Literal["dengue", "zika", "chikungunya"]) -> str:
    assert disease.lower() in DISEASES, (
        f"Unknown disease '{disease}'. Options: {DISEASES}"
    )
    return disease


def validate_adm_level(adm_level: int) -> int:
    assert adm_level in ADM_LEVELS, (
        f"Unknown adm_level {adm_level}. Options {ADM_LEVELS}"
    )
    return adm_level


def validate_time_resolution(time_resolution: str) -> str:
    assert time_resolution in TIME_RESOLUTIONS, (
        f"Unkown time_resolution {time_resolution}. ",
        f"Options: {TIME_RESOLUTIONS}"
    )
    return time_resolution


def validate_temporal(temporal: bool) -> bool:
    return temporal


def validate_spatial(spatial: bool) -> bool:
    return spatial


def validate_categorical(categorical: bool) -> bool:
    return categorical


def validate_commit(commit: str) -> str:
    err = "Invalid GitHub commit hash"
    assert len(commit) == 40, err
    assert all(c in ascii_lowercase + digits for c in commit), err
    return commit


def validate_date(date: str | dt.date) -> str:
    error = "Incorrect date format. Format: YYYY-MM-DD \n%s"
    try:
        dt_date = dt.date.fromisoformat(str(date))
    except Exception as err:
        raise ValidationError(error % err)

    assert dt_date >= dt.date(2010, 1, 1), "date is too old"
    assert dt_date <= dt.datetime.now().date(), "date is in the future"

    return str(date)


def validate_prediction_data(data: str | list | pd.DataFrame) -> pd.DataFrame:
    if isinstance(data, list):
        df = pd.DataFrame(data)
    elif isinstance(data, str):
        try:
            df = pd.DataFrame(json.loads(data))
        except json.decoder.JSONDecodeError:
            raise ValueError(
                "`data` object must be JSON serializable or a DataFrame"
            )
    elif isinstance(data, pd.DataFrame):
        df = data
    else:
        raise ValueError(f"Invalid `data` type {type(data)}")

    if df.empty:
        logging.warning("Empty data")
        return data

    assert set(df.columns) == set(PREDICTION_DATA_COLUMNS), (
        f"Incorrect data columns. Expecting: {PREDICTION_DATA_COLUMNS}; "
        f"Missing {set(PREDICTION_DATA_COLUMNS).difference(set(df.columns))}"
    )
    # TODO: Include more checks
    return data


def validate_tags(tags: list[int]) -> list[int]:
    # TODO:
    return tags

# fmt: off
def validate_uf(
    uf: Literal[
        "AC", "AL", "AP", "AM", "BA", "CE", "ES", "GO", "MA", "MT", "MS", "MG",
        "PA", "PB", "PR", "PE", "PI", "RJ", "RN", "RS", "RO", "RR", "SC", "SP",
        "SE", "TO", "DF"
]) -> str:
    # fmt: on
    assert uf.upper() in UF_CODES, f"Unknown UF {uf}. Example: 'SP'"
    return uf.upper()


def validate_geocode(geocode: int) -> int:
    err = "Invalid municipality geocode %s. Example: 3304557"
    assert len(str(geocode)) == 7, err % geocode
    assert int(str(geocode)[:2]) in UF_CODES.values(), err % geocode
    return geocode



from urllib.parse import urljoin
from typing import Literal, Optional, Any, Dict, AnyStr, List

import json
import requests
import nest_asyncio
import pandas as pd
from pydantic import BaseModel, ConfigDict

from mosqlient import types
from mosqlient.client import Client
from mosqlient.errors import ClientError, ModelPostError, PredictionPostError
from mosqlient.requests import get_all_sync
from mosqlient._utils import parse_params
from mosqlient._config import get_api_url
from mosqlient.registry import schema


nest_asyncio.apply()


class Base(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        protected_namespaces=()
    )

    def __str__(self):
        return self.json()

    def json(self):
        return self._schema.json()


class User(Base):
    _schema: schema.UserSchema

    def __init__(
        self,
        name: types.AuthorName,
        username: types.AuthorUserName,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._schema = schema.UserSchema(
            name=name,
            username=username
        )

    @property
    def name(self) -> types.AuthorName:
        return self._schema.name

    @property
    def username(self) -> types.AuthorUserName:
        return self._schema.username


class Author(Base):
    user: User
    _schema: schema.AuthorSchema

    def __init__(
        self,
        user: User | dict,
        institution: types.AuthorInstitution,
        **kwargs
    ):

        if isinstance(user, dict):
            user = User(**user)

        kwargs['user'] = user

        super().__init__(**kwargs)
        self.user = user

        self._schema = schema.AuthorSchema(
            user=user._schema,
            institution=institution
        )

    def __repr__(self) -> str:
        return self._schema.user.name

    @property
    def institution(self) -> types.AuthorInstitution:
        return self._schema.institution

    @classmethod
    def get(
        cls,
        name: Optional[types.AuthorName] = None,
        institution: Optional[types.AuthorInstitution] = None,
        username: Optional[types.AuthorUserName] = None,
        **kwargs
    ):
        timeout = kwargs["timeout"] if "timeout" in kwargs else 300

        params = {
            "name": name,
            "institution": institution,
            "username": username
        }
        params = parse_params(**params)

        return [
            Author(**m) for m in get_all_sync(
                app="registry",
                endpoint="authors",
                params=params,
                pagination=False,
                timeout=timeout
            )
        ]


class ImplementationLanguage(Base):
    _schema: schema.ImplementationLanguageSchema

    def __init__(self, language: types.ImplementationLanguage, **kwargs):
        super().__init__(**kwargs)
        if isinstance(language, dict):
            self._schema = schema.ImplementationLanguageSchema(
                language=language['language']
            )
        elif isinstance(language, str):
            self._schema = schema.ImplementationLanguageSchema(
                language=language
            )
        else:
            raise ValueError(
                "`language` must be a str or a dict with {'language': `language`}"
            )

    def __repr__(self) -> str:
        return self._schema.language

    def __str__(self) -> str:
        return self._schema.language

    @property
    def language(self):
        return self._schema.language


class Model(Base):
    client: Optional[Client] = None
    author: Author
    implementation_language: ImplementationLanguage
    _schema: schema.ModelSchema

    def __init__(
        self,
        name: types.Name,
        description: types.Description,
        author: Author | dict,
        repository: types.Repository,
        implementation_language: types.ImplementationLanguage,
        disease: types.Disease,
        categorical: types.Categorical,
        spatial: types.Spatial,
        temporal: types.Temporal,
        ADM_level: types.ADMLevel,
        time_resolution: types.TimeResolution,
        id: Optional[types.ID] = None,
        **kwargs
    ):

        if isinstance(author, dict):
            author = Author(
                user=author['user'],
                institution=author['institution']
            )
        kwargs['author'] = author

        language = ImplementationLanguage(
            language=implementation_language
        )
        kwargs['implementation_language'] = language

        super().__init__(**kwargs)

        self.author = author
        self.implementation_language = language

        self._schema = schema.ModelSchema(
            id=id,
            name=name,
            description=description,
            author=author._schema,
            repository=repository,
            implementation_language=language._schema,
            disease=disease,
            categorical=categorical,
            spatial=spatial,
            temporal=temporal,
            ADM_level=ADM_level,
            time_resolution=time_resolution,
        )

    def __repr__(self) -> str:
        return self.name

    @property
    def id(self) -> types.ID | None:
        return self._schema.id

    @property
    def name(self) -> types.Name:
        return self._schema.name

    @property
    def description(self) -> types.Description:
        return self._schema.description

    @property
    def repository(self) -> types.Repository:
        return self._schema.repository

    @property
    def disease(self) -> types.Disease:
        return self._schema.disease

    @property
    def categorical(self) -> types.Categorical:
        return self._schema.categorical

    @property
    def spatial(self) -> types.Spatial:
        return self._schema.spatial

    @property
    def temporal(self) -> types.Temporal:
        return self._schema.temporal

    @property
    def ADM_level(self) -> types.ADMLevel:
        return self._schema.ADM_level

    @property
    def time_resolution(self) -> types.TimeResolution:
        return self._schema.time_resolution

    @classmethod
    def get(cls, **kwargs):
        """
        https://api.mosqlimate.org/docs/registry/GET/models/
        """
        timeout = kwargs["timeout"] if "timeout" in kwargs else 300

        ModelGETParams(**kwargs)
        params = parse_params(**kwargs)

        return [
            Model(**m) for m in get_all_sync(
                app="registry",
                endpoint="models",
                params=params,
                pagination=True,
                timeout=timeout
            )
        ]

    def post(self, **kwargs):
        """
        https://api.mosqlimate.org/docs/registry/POST/models/
        """
        timeout = kwargs["timeout"] if "timeout" in kwargs else 10

        if self.client is None:
            raise ClientError(
                "A Client instance must be provided, please instantiate Model "
                "passing your Mosqlimate's credentials. For more info about "
                "retrieving or inserting data from Mosqlimate, please see the "
                "API Documentation"
            )

        params = {
            "name": self.name,
            "description": self.description,
            "repository": self.repository,
            "implementation_language": str(self.implementation_language),
            "disease": self.disease,
            "temporal": self.temporal,
            "spatial": self.spatial,
            "categorical": self.categorical,
            "ADM_level": self.ADM_level,
            "time_resolution": self.time_resolution,
        }

        url = urljoin(get_api_url(), "/".join(("registry", "models")) + "/")
        headers = {"X-UID-Key": self.client.X_UID_KEY}

        resp = requests.post(
            url,
            json=params,
            headers=headers,
            timeout=timeout
        )

        if resp.status_code != 201:
            raise ModelPostError(
                "POST request returned status code "
                f"{resp.status_code} \n {resp.reason} \n {resp.json()}"
            )

        return resp

    def update(
        self,
        id: int,
        name: Optional[str] = None,
        description: Optional[str] = None,
        repository: Optional[str] = None,
        implementation_language: Optional[str] = None,
        disease: Optional[Literal["dengue", "chikungunya", "zika"]] = None,
        temporal: Optional[bool] = None,
        spatial: Optional[bool] = None,
        categorical: Optional[bool] = None,
        adm_level: Optional[Literal[0, 1, 2, 3]] = None,
        # fmt: off
        time_resolution: Optional[Literal["day", "week", "month", "year"]] = None,
        # fmt: on
        **kwargs
    ):
        """
        https://github.com/Mosqlimate-project/Data-platform/blob/main/src/registry/api.py#L258
        """
        timeout = kwargs["timeout"] if "timeout" in kwargs else 10

        if self.client is None:
            raise ClientError(
                "A Client instance must be provided, please instantiate Model "
                "passing your Mosqlimate's credentials. For more infor about "
                "retrieving or inserting data from Mosqlimate, please see the "
                "API Documentation"
            )

        params = {
            "name": name,
            "description": description,
            "repository": repository,
            "implementation_language": implementation_language,
            "disease": disease,
            "temporal": temporal,
            "spatial": spatial,
            "categorical": categorical,
            "ADM_level": adm_level,
            "time_resolution": time_resolution
        }

        ModelPUTParams(
            id=id,
            name=name,
            description=description,
            repository=repository,
            implementation_language=implementation_language,
            disease=disease,
            temporal=temporal,
            spatial=spatial,
            categorical=categorical,
            ADM_level=adm_level,
            time_resolution=time_resolution,
        )

        url = urljoin(
            get_api_url(), "/".join(("registry", "models")) + f"/{id}"
        )
        headers = {"X-UID-Key": self.client.X_UID_KEY}
        resp = requests.put(url, json=params, headers=headers, timeout=timeout)

        return resp


class ModelGETParams(Base):
    # https://github.com/Mosqlimate-project/Data-platform/blob/main/src/registry/schema.py#L43
    id: Optional[types.ID] = None
    name: Optional[types.Name] = None
    author_name: Optional[types.AuthorName] = None
    author_username: Optional[types.AuthorUserName] = None
    author_institution: Optional[types.AuthorInstitution] = None
    repository: Optional[types.Repository] = None
    implementation_language: Optional[types.ImplementationLanguage] = None
    disease: Optional[types.Disease] = None
    ADM_level: Optional[types.ADMLevel] = None
    temporal: Optional[types.Temporal] = None
    spatial: Optional[types.Spatial] = None
    categorical: Optional[types.Categorical] = None
    time_resolution: Optional[types.TimeResolution] = None
    tags: Optional[types.Tags] = None


class ModelPUTParams(Base):
    id: types.ID
    name: Optional[types.Name] = None
    description: Optional[types.Description] = None
    repository: Optional[types.Repository] = None
    implementation_language: Optional[types.ImplementationLanguage] = None
    disease: Optional[types.Disease] = None
    ADM_level: Optional[types.ADMLevel] = None
    temporal: Optional[types.Temporal] = None
    spatial: Optional[types.Spatial] = None
    categorical: Optional[types.Categorical] = None
    time_resolution: Optional[types.TimeResolution] = None


class Prediction(Base):
    client: Optional[Client] = None
    model: Model
    _schema: schema.PredictionSchema

    def __init__(
        self,
        model: Model | dict,
        description: types.Description,
        commit: types.Commit,
        predict_date: types.Date,
        data: types.PredictionData,
        id: Optional[types.ID] = None,
        **kwargs
    ):
        if isinstance(model, dict):
            model = Model(**model)

        kwargs['model'] = model

        super().__init__(**kwargs)

        if isinstance(data, str):
            try:
                _data = json.loads(data)
            except json.decoder.JSONDecodeError:
                raise ValueError("str `data` must be JSON serializable")
            _data = [
                schema.PredictionDataRowSchema(**d)
                for d in _data
            ]
        elif isinstance(data, pd.DataFrame):
            _data = [
                schema.PredictionDataRowSchema(**d)
                for d in data.to_dict(orient="records")
            ]
        elif isinstance(data, list):
            _data = [
                schema.PredictionDataRowSchema(**d)
                for d in data
            ]
        else:
            raise ValueError(
                "`data` must be rather a DataFrame, a JSON str or a list of" +
                " dictionaries"
            )

        self.model = model
        self._schema = schema.PredictionSchema(
            id=id,
            model=model._schema,
            description=description,
            commit=commit,
            predict_date=predict_date,
            data=_data
        )

    def __repr__(self) -> str:
        return f"Prediction <{self.id}>"

    @property
    def id(self) -> types.ID | None:
        return self._schema.id

    @property
    def description(self) -> types.Description:
        return self._schema.description

    @property
    def commit(self) -> types.Commit:
        return self._schema.commit

    @property
    def predict_date(self) -> types.Date:
        return self._schema.predict_date

    @property
    def data(self) -> List[Dict[AnyStr, Any]]:
        return [row.dict() for row in self._schema.data]

    def to_dataframe(self) -> pd.DataFrame:
        return pd.DataFrame(self.data)

    @classmethod
    def get(cls, **kwargs):
        """
        [DOCUMENTATION](https://api.mosqlimate.org/docs/registry/GET/predictions/)

        SEARCH PARAMETERS
        All parameters are Optional, use them to filter the Predictions in the
        result

        id [int]: Search by Prediction ID
        model_id [int]: Search by Model ID
        model_name [str]: Search by Model Name
        model_ADM_level [int]: Search by ADM Level
        model_time_resolution [str]: Search by Time Resolution
        model_disease [str]: Search by Disease
        author_name [str]: Search by Author Name
        author_username [str]: Search by Author Username
        author_institution [str]: Search by Author Institution
        repository [str]: Search by GitHub repository
        implementation_language [str]: Search by Implementation Language
        temporal [bool]: Search by Temporal Models
        spatial [bool]: Search by Spatial Models
        categorical [bool]: Search by Categorical Models
        commit [str]: Search by Git Commit
        predict_date [str]: Search by prediction date. Format: YYYY-MM-DD
        start [str]: Search by prediction start date. Format: YYYY-MM-DD
        end [str]: Search by prediction end date. Format: YYYY-MM-DD
        """
        timeout = kwargs["timeout"] if "timeout" in kwargs else 300

        PredictionGETParams(**kwargs)
        params = parse_params(**kwargs)
        data = get_all_sync(
            app="registry",
            endpoint="predictions",
            params=params,
            pagination=True,
            timeout=timeout,
        )
        return [Prediction(**p) for p in data]

    def post(self, **kwargs):
        timeout = kwargs["timeout"] if "timeout" in kwargs else 30

        if self.id is not None:
            raise PredictionPostError("The Prediction already has an ID")

        if self.client is None:
            raise ClientError(
                "A Client instance must be provided, please instantiate Model "
                "passing your Mosqlimate's credentials. For more info about "
                "retrieving or inserting data from Mosqlimate, please see the "
                "API Documentation"
            )

        params = {
            "model": self.model.id,
            "description": self.description,
            "commit": self.commit,
            "predict_date": self.predict_date,
            "prediction": json.dumps(self.data),
        }

        url = urljoin(
            get_api_url(),
            "/".join(("registry", "predictions")) + "/"
        )
        headers = {"X-UID-Key": self.client.X_UID_KEY}

        resp = requests.post(
            url,
            json=params,
            headers=headers,
            timeout=timeout
        )

        if str(resp.status_code).startswith("5"):
            raise ClientError(
                f"{resp.status_code}: {resp.reason}. " +
                "This error is from the API Server, please try again and " +
                "contact the moderation if this error persists"
            )

        if resp.status_code != 201:
            raise PredictionPostError(
                "POST request returned status code " +
                f"{resp.status_code}: {resp.reason} \n {resp.json()}"
            )

        # TODO: Return a Prediction object retrieving it from the API
        return resp


class PredictionGETParams(Base):
    id: Optional[types.ID] = None
    model: Optional[types.ID] = None
    model_name: Optional[types.Name] = None
    model_ADM_level: Optional[types.ADMLevel] = None
    model_time_resolution: Optional[types.TimeResolution] = None
    model_disease: Optional[types.Disease] = None
    author_name: Optional[types.AuthorName] = None
    author_username: Optional[types.AuthorUserName] = None
    author_institution: Optional[types.AuthorInstitution] = None
    repository: Optional[types.Repository] = None
    implementation_language: Optional[types.ImplementationLanguage] = None
    temporal: Optional[types.Temporal] = None
    spatial: Optional[types.Spatial] = None
    categorical: Optional[types.Categorical] = None
    commit: Optional[types.Commit] = None
    predict_date: Optional[types.Date] = None
    start: Optional[types.Date] = None
    end: Optional[types.Date] = None


class PredictionPUTParams(Base):
    model: types.ID
    description: Optional[types.Description] = None
    commit: Optional[types.Commit] = None
    predict_date: Optional[types.Date] = None
    prediction: Optional[types.PredictionData] = None



__all__ = [
    "get_all_models",
    "get_models",
    "get_model_by_id",
    "get_models_by_author_name",
    "get_models_by_author_username",
    "get_models_by_author_institution",
    "get_models_by_repository",
    "get_models_by_implementation_language",
    "get_models_by_disease",
    "get_models_by_adm_level",
    "get_models_by_time_resolution",
    "get_models_by_tag_ids"
]

from typing import Optional, List, Union

from .models import Model


def get_all_models() -> list[dict]:
    return Model.get()


def get_models(
    id: Optional[int] = None,
    name: Optional[str] = None,
    author_name: Optional[str] = None,
    author_username: Optional[str] = None,
    author_institution: Optional[str] = None,
    repository: Optional[str] = None,
    implementation_language: Optional[str] = None,
    disease: Optional[str] = None,
    ADM_level: Optional[int] = None,
    temporal: Optional[bool] = None,
    spatial: Optional[bool] = None,
    categorical: Optional[bool] = None,
    time_resolution: Optional[str] = None,
    tags: Optional[List[int]] = None,
) -> Union[dict, list[dict]]:
    params = {
        "id": id,
        "name": name,
        "author_name": author_name,
        "author_username": author_username,
        "author_institution": author_institution,
        "repository": repository,
        "implementation_language": implementation_language,
        "disease": disease,
        "ADM_level": ADM_level,
        "temporal": temporal,
        "spatial": spatial,
        "categorical": categorical,
        "time_resolution": time_resolution,
        "tags": tags,
    }
    res = Model.get(**params)
    return res[0] if len(res) == 1 else res


def get_model_by_id(id: int) -> Union[dict, list]:
    res = Model.get(id=id)
    return res[0] if len(res) == 1 else res


def get_models_by_author_name(author_name: str) -> Union[dict, list[dict]]:
    res = Model.get(author_name=author_name)
    return res[0] if len(res) == 1 else res


def get_models_by_author_username(
    author_username: str
) -> Union[dict, list[dict]]:
    res = Model.get(author_username=author_username)
    return res[0] if len(res) == 1 else res


def get_models_by_author_institution(
    author_institution: str
) -> Union[dict, list[dict]]:
    res = Model.get(author_institution=author_institution)
    return res[0] if len(res) == 1 else res


def get_models_by_repository(repository: str) -> Union[dict, list[dict]]:
    res = Model.get(repository=repository)
    return res[0] if len(res) == 1 else res


def get_models_by_implementation_language(
    implementation_language: str
) -> Union[dict, list[dict]]:
    res = Model.get(implementation_language=implementation_language)
    return res[0] if len(res) == 1 else res


def get_models_by_disease(disease: str) -> Union[dict, list[dict]]:
    res = Model.get(disease=disease)
    return res[0] if len(res) == 1 else res


def get_models_by_adm_level(adm_level: int) -> Union[dict, list[dict]]:
    res = Model.get(adm_level=adm_level)
    return res[0] if len(res) == 1 else res


def get_models_by_time_resolution(
    time_resolution: int
) -> Union[dict, list[dict]]:
    res = Model.get(time_resolution=time_resolution)
    return res[0] if len(res) == 1 else res


def get_models_by_tag_ids(tags_ids: list[int]) -> Union[dict, list[dict]]:
    res = Model.get(tags=tags_ids)
    return res[0] if len(res) == 1 else res



import asyncio
from itertools import chain
from typing import AnyStr, List, Any
from urllib.parse import urljoin

import aiohttp
import requests

from mosqlient.types import APP
from mosqlient._config import get_api_url


def get(
    app: APP,
    endpoint: AnyStr,
    params: dict[str, str | int | float],
    pagination: bool = True,
    timeout: int = 300,
) -> requests.models.Response:
    if pagination and ("page" not in params or "per_page" not in params):
        raise ValueError(
            "'page' and 'per_page' parameters are required to requests",
            " with pagination"
        )

    if not endpoint:
        raise ValueError("endpoint is required")

    url = urljoin(get_api_url(), "/".join((str(app), str(endpoint)))) + "/?"

    return requests.get(url, params, timeout=timeout)


async def aget(
    session: aiohttp.ClientSession,
    url: str,
    params: dict[str, str | int | float],
    timeout: int = 300,
    retries: int = 3
) -> Any:
    try:
        if retries < 0:
            raise aiohttp.ClientConnectionError("Too many attempts")
        async with session.get(url, params=params, timeout=timeout) as res:
            if res.status == 200:
                return await res.json()
            if str(res.status).startswith("4"):
                raise aiohttp.ClientConnectionError(
                    f"Response status: {res.status}. Reason: {res.reason}"
                )
            if retries == 0:
                raise aiohttp.ClientConnectionError(
                    f"Response status: {res.status}. Reason: {res.reason}"
                )
            await asyncio.sleep(10 / (retries + 1))
            return await aget(session, url, params, timeout, retries - 1)
    except aiohttp.ServerTimeoutError:
        await asyncio.sleep(8 / (retries + 1))
        return await aget(session, url, params, timeout, retries - 1)
    raise aiohttp.ClientConnectionError("Invalid request")


async def get_all(
    app: APP,
    endpoint: AnyStr,
    params: dict[str, str | int | float],
    timeout: int = 300,
    pagination: bool = False,
    _max_per_page: int = 300,
) -> List[dict]:
    if pagination:
        params["page"] = 1
        params["per_page"] = _max_per_page

    url = urljoin(get_api_url(), "/".join((str(app), str(endpoint)))) + "/?"

    async with aiohttp.ClientSession() as session:
        first_page = await aget(session, url, params)

    if not pagination:
        return first_page

    if not first_page:
        return first_page

    total_pages = first_page["pagination"]["total_pages"]

    if total_pages == 1:
        return first_page["items"]

    async with aiohttp.ClientSession(
        timeout=aiohttp.ClientTimeout(total=timeout)
    ) as session:
        tasks = []
        for page in range(2, total_pages + 1):
            params_c = params.copy()
            params_c["page"] = page
            task = asyncio.create_task(aget(session, url, params_c))
            tasks.append(task)
        results = await asyncio.gather(*tasks)

    if results:
        results.insert(0, first_page)

    res = list(chain.from_iterable(
        result["items"] for result in results if result is not None
    ))

    return res


def get_all_sync(
    app: APP,
    endpoint: AnyStr,
    params: dict[str, Any],
    timeout: int = 300,
    pagination: bool = False,
    _max_per_page: int = 300,
):
    async def fetch_all():
        return await get_all(
            app=app,
            endpoint=endpoint,
            params=params,
            timeout=timeout,
            pagination=pagination
        )

    if asyncio.get_event_loop().is_running():
        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(fetch_all())
        return loop.run_until_complete(future)
    return asyncio.run(fetch_all())



from datetime import date
from typing import Optional
from pydantic import BaseModel, ConfigDict

from mosqlient import types
from mosqlient.requests import get_all_sync
from mosqlient._utils import parse_params


class Climate(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def get(cls, start: str | date, end: str | date, **kwargs):
        timeout = kwargs["timeout"] if "timeout" in kwargs else 300

        kwargs["start"] = start
        kwargs["end"] = end
        params = parse_params(**kwargs)
        ClimateGETParams(**kwargs)

        return get_all_sync(
            app="datastore",
            endpoint="climate",
            params=params,
            pagination=True,
            timeout=timeout
        )


class ClimateGETParams(BaseModel):
    start: Optional[types.Date] = None
    end: Optional[types.Date] = None
    geocode: Optional[types.Geocode] = None
    uf: Optional[types.UF] = None



__all__ = [
    "get_all_predictions",
    "get_predictions",
    "get_prediction_by_id",
    "get_predictions_by_model_id",
    "get_predictions_by_model_name",
    "get_predictions_by_adm_level",
    "get_predictions_by_time_resolution",
    "get_predictions_by_disease",
    "get_predictions_by_author_name",
    "get_predictions_by_author_username",
    "get_predictions_by_author_institution",
    "get_predictions_by_repository",
    "get_predictions_by_implementation_language",
    "get_predictions_by_predict_date",
    "get_predictions_between"
]

from datetime import date

from typing import Optional, Union, List

from .models import Prediction


def get_all_predictions() -> list[dict]:
    return Prediction.get()


def get_predictions(
    id: Optional[int] = None,
    model_id: Optional[int] = None,
    model_name: Optional[str] = None,
    model_ADM_level: Optional[int] = None,
    model_time_resolution: Optional[str] = None,
    model_disease: Optional[str] = None,
    author_name: Optional[str] = None,
    author_username: Optional[str] = None,
    author_institution: Optional[str] = None,
    repository: Optional[str] = None,
    implementation_language: Optional[str] = None,
    temporal: Optional[bool] = None,
    spatial: Optional[bool] = None,
    categorical: Optional[bool] = None,
    commit: Optional[str] = None,
    predict_date: Optional[date] = None,
    start: Optional[date] = None,
    end: Optional[date] = None,
) -> list[dict] | dict:

    params = {
        "id": id,
        "model_id": model_id,
        "model_name": model_name,
        "model_ADM_level": model_ADM_level,
        "model_time_resolution": model_time_resolution,
        "model_disease": model_disease,
        "author_name": author_name,
        "author_username": author_username,
        "author_institution": author_institution,
        "repository": repository,
        "implementation_language": implementation_language,
        "temporal": temporal,
        "spatial": spatial,
        "categorical": categorical,
        "commit": commit,
        "predict_date": str(predict_date) if predict_date is not None else None,
        "start": str(start) if start is not None else None,
        "end": str(end) if end is not None else None,
    }

    res = Prediction.get(**params)
    return res[0] if len(res) == 1 else res


def get_prediction_by_id(id: int) -> Prediction:
    res = Prediction.get(id=id)
    return res[0] if len(res) == 1 else res


def get_predictions_by_model_id(model_id: int) -> List[Prediction]:
    return Prediction.get(model_id=model_id)


def get_predictions_by_model_name(model_name: str) -> List[Prediction]:
    return Prediction.get(model_name=model_name)


def get_predictions_by_adm_level(adm_level: int) -> List[Prediction]:
    return Prediction.get(adm_level=adm_level)


def get_predictions_by_time_resolution(
    time_resolution: str
) -> List[Prediction]:
    return Prediction.get(time_resolution=time_resolution)


def get_predictions_by_disease(disease: str) -> List[Prediction]:
    return Prediction.get(disease=disease)


def get_predictions_by_author_name(author_name: str) -> List[Prediction]:
    return Prediction.get(author_name=author_name)


def get_predictions_by_author_username(
    author_username: str
) -> List[Prediction]:
    return Prediction.get(author_username=author_username)


def get_predictions_by_author_institution(
    author_institution: str
) -> List[Prediction]:
    return Prediction.get(author_institution=author_institution)


def get_predictions_by_repository(
    repository: str
) -> List[Prediction]:
    return Prediction.get(repository=repository)


def get_predictions_by_implementation_language(
    implementation_language: str
) -> List[Prediction]:
    return Prediction.get(implementation_language=implementation_language)


def get_predictions_by_predict_date(predict_date: date) -> List[Prediction]:
    return Prediction.get(predict_date=str(predict_date))


def get_predictions_between(start: date, end: date) -> List[Prediction]:
    return Prediction.get(start=str(start), end=str(end))



__all__ = ["upload_prediction"]

import requests
from datetime import date

import pandas as pd

from mosqlient import Client
from .models import Prediction, Model


def upload_prediction(
    model_id: int,
    description: str,
    commit: str,
    predict_date: str | date,
    prediction: str | list[dict] | pd.DataFrame,
    api_key: str,
    **kwargs
) -> requests.Response:
    client = Client(x_uid_key=api_key)
    model = Model.get(id=model_id)[0]
    return Prediction(
        client=client,
        model=model,
        description=description,
        commit=commit,
        predict_date=predict_date,
        data=prediction,
    ).post(**kwargs)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from pmdarima.arima import auto_arima
from pmdarima import preprocessing as ppc
from mosqlient.datastore import Infodengue
from datetime import date, datetime, timedelta

def get_next_n_weeks(ini_date: str, next_days: int) -> list:
    """
    Return a list of dates with the {next_weeks} weeks after ini_date.
    This function was designed to generate the dates of the forecast
    models.
    Parameters
    ----------
    ini_date : str
        Initial date.
    next_weeks : int
        Number of weeks to be included in the list after the date in
        ini_date.
    Returns
    -------
    list
        A list with the dates computed.
    """

    next_dates = []

    a = datetime.strptime(ini_date, "%Y-%m-%d")

    for i in np.arange(1, next_days + 1):
        d_i = datetime.strftime(a + timedelta(days=int(i * 7)), "%Y-%m-%d")

        next_dates.append(datetime.strptime(d_i, "%Y-%m-%d").date())

    return next_dates

def get_prediction_dataframe(preds, date, boxcox) ->pd.DataFrame:
    """
    Function to organize the predictions of the ARIMA model in a pandas DataFrame.

    Parameters
    ----------
    horizon: int
        The number of weeks forecasted by the model
    end_date: str
        Last week of the out of the sample evaluation. The first week is after the last training observation.
    plot: bool
        If true the plot of the model out of the sample is returned
    """

    df_preds = pd.DataFrame()

    df_preds['dates'] = date

    try:
        df_preds['preds'] = preds[0].values

    except:
        df_preds['preds'] = preds[0]

    df_preds.loc[:, ['lower', 'upper']] = preds[1]

    if df_preds['preds'].values[0] == 0:
        df_preds = df_preds.iloc[1:]

    df_preds['preds'] =  boxcox.inverse_transform(df_preds['preds'])[0]
    df_preds['lower'] =  boxcox.inverse_transform(df_preds['lower'])[0]
    df_preds['upper'] =  boxcox.inverse_transform(df_preds['upper'])[0]

    return df_preds

def plot_predictions(df_preds:pd.DataFrame, title:str = '') -> None:
    """
    Function to plot the predictions of the model.

    Parameters
    ----------
    df_preds: pd.DataFrame
        Dataframe with the columns: ['dates', 'data', 'preds', 'lower', 'upper'].
    title: str
        Title of the plot.
    """

    fig,ax = plt.subplots(1, figsize = (6,4))

    ax.plot(df_preds.dates, df_preds.data, color = 'black', label = 'Data')

    ax.plot(df_preds.dates, df_preds.preds, color = 'tab:orange', label = 'ARIMA')

    ax.fill_between(df_preds.dates, df_preds.lower, df_preds.upper, color = 'tab:orange', alpha = 0.3)

    ax.legend()

    ax.grid()

    ax.set_title(title)

    ax.xaxis.set_major_formatter(mdates.DateFormatter('%b\n%y'))

    ax.set_xlabel('Date')

    ax.set_ylabel('New cases')



def plot_forecast(df_for: pd.DataFrame, df_train: pd.DataFrame, last_obs:int) -> None:
    """
    Function to plot the forecast of the model.

    Parameters
    ----------
    df_for: pd.DataFrame
        Dataframe with the forecast results, with the columns: ['dates', 'preds', 'lower', 'upper']
    df_preds: pd.DataFrame
        Dataframe with the columns: ['data'] and a datetime index.
    last_obs: int
        Number of previous observations of the data included.
    """

    df_train = df_train.tail(last_obs)

    fig,ax = plt.subplots(1, figsize = (6,4))

    ax.plot(df_train.index, df_train.data, color = 'black', label = 'Data')

    ax.plot(df_for.dates, df_for.preds, color = 'tab:red', label = 'Forecast')

    ax.fill_between(df_for.dates, df_for.lower, df_for.upper, color = 'tab:red', alpha = 0.3)

    ax.plot([df_train.index[-1], df_for.dates[0]], [df_train[f'data'].values[-1], df_for.preds.values[0]], ls = '--', color = 'black')

    ax.legend()

    ax.grid()

    ax.set_title('Forecast ARIMA')

    ax.set_xlabel('Date')

    ax.set_ylabel('New cases')

    ax.xaxis.set_major_formatter(mdates.DateFormatter('%b-%d\n%Y'))


class InvalidDataFrameError(Exception):
    """Custom exception for invalid DataFrame."""
    pass

class Arima:
    """
    A class to implement a ARIMA model as baseline for forecast cases in some city.

    Attributes
    ----------
    df : pd.DataFrame
        A pandas dataframe with the columns y and a datetime index

    Methods
    -------
    train():
        Train the model.
    predict_out_of_sample():
        Predictions of the model in sample.
    predict_out_of_sample():
        Predictions of the model out of sample.
    forecast():
        Forecast models
    """

    def __init__(self, df:pd.DataFrame):
        """
        Constructs all the necessary attributes for the Arima object.

        Parameters
        ----------
            df : pd.DataFrame
            A pandas dataframe with the column y and a datetime index

        """
        if not pd.api.types.is_datetime64_any_dtype(df.index):
            raise InvalidDataFrameError("The DataFrame's index is not of datetime type.")

        if df.shape[1] != 1:
            raise InvalidDataFrameError("The DataFrame must have one single column.")

        if df.columns[0] != 'y':
            raise InvalidDataFrameError("The column must be named `y`.")

        df['y']= df['y'].astype(float)

        self.df = df

    def train(self, train_ini_date:str, train_end_date:str):
        """
        Train the ARIMA model

        Parameters
        ----------
            train_ini_date: str
                Initial date for model training
            train_end_date: str
                End date for model training
        """

        df_train = self.df.copy()

        df_train = df_train.loc[(df_train.index >= pd.to_datetime(train_ini_date)) & (df_train.index <= pd.to_datetime(train_end_date))]

        self.df_train = df_train

        boxcox = ppc.BoxCoxEndogTransformer().fit(df_train.y)

        self.boxcox = boxcox

        df_train.loc[:, 'y'] = boxcox.transform(df_train.y)[0]

        model = auto_arima(df_train.y,
                    seasonal=False,
                    trace=True,
                    maxiter = 100,
                    error_action='ignore',
                    information_criterion = 'aic',suppress_warnings=True,
                    stepwise = True,
                   )


        model.fit(df_train.y)

        self.model = model

        return model

    def predict_in_sample(self, plot:bool = True) -> pd.DataFrame:
        """
        Returns the model performance in the sample.

        Parameters
        ----------
            plot: bool
                If true the plot of the model in the sample is returned

        """

        preds_in_sample = self.model.predict_in_sample(return_conf_int=True)

        df_train = self.df_train.copy()

        df_in_sample = get_prediction_dataframe(preds_in_sample, df_train.index, self.boxcox)

        df_in_sample = df_in_sample.merge(df_train, left_on = 'dates', right_index = True)

        df_in_sample = df_in_sample.rename(columns = {'y': 'data'})

        df_in_sample['data'] = self.boxcox.inverse_transform(df_in_sample['data'])[0]

        if plot:

            plot_predictions(df_in_sample, title = 'In sample predictions')

        return df_in_sample


    def predict_out_of_sample(self, horizon:int, end_date:str, plot = True)->pd.DataFrame:
        """
        Returns the model performance out of the sample.
        The predictions are returned by windows of {horizon} observations. After each window
        the model is updated with the data of the last observations forecasted.

        Parameters
        ----------
            horizon: int
                The number of observations forecasted by the model
            end_date: str
                Last week of the out of sample evaluation. The first week is after the last training observation.
            plot: bool
                If true the plot of the model out of the sample is returned
        """


        df = self.df.copy()

        df.loc[:, 'y'] = self.boxcox.transform(df.y)[0]

        model = self.model

        preds = model.predict(horizon, return_conf_int = True)

        dates = get_next_n_weeks(self.df_train.index[-1].strftime( "%Y-%m-%d"), horizon)

        df_preds = get_prediction_dataframe(preds, dates, self.boxcox)

        while ( pd.Timestamp(dates[-1]) < pd.to_datetime(end_date)):

            preds = model.update(df.loc[dates[0] : dates[-1]]).predict(horizon, return_conf_int = True)

            dates = get_next_n_weeks(dates[-1].strftime( "%Y-%m-%d"), horizon)

            df_preds = pd.concat([df_preds, get_prediction_dataframe(preds, dates, self.boxcox)])

        df_preds.dates = pd.to_datetime(df_preds.dates)

        df_preds = df_preds.merge(df, left_on = 'dates', right_index = True)

        df_preds = df_preds.loc[df_preds.dates <= end_date]

        df_preds = df_preds.dropna()

        df_preds.loc[:, 'y'] = self.boxcox.inverse_transform(df_preds['y'])[0]

        df_preds = df_preds.rename(columns = {'y': 'data'})

        if plot:

            plot_predictions(df_preds, title = 'Out of sample predictions')

        return df_preds


    def forecast(self, horizon:int, plot:bool, last_obs:int) -> pd.DataFrame:
        """
        Returns the forecast of the model.
        Before applying this method is necessary to call the `train()` method.
        The `forecast()` method will forecast {horizon} observations ahead of the last observation
        used to train the model in the `train()` method.

        Parameters
        ----------
            horizon: int
                The number of observations forecasted by the model
            plot: bool
                If true return a figure with the forecasted values.
            last_obs: bool
                The number of last observations plotted in the figure
        """

        df_train = self.df_train.copy()

        df_train = df_train.rename(columns = {'y': 'data'})

        df_train['data'] = self.boxcox.inverse_transform(df_train['data'])[0]

        model = self.model

        dates = get_next_n_weeks(df_train.index[-1].strftime( "%Y-%m-%d"), horizon)

        preds = model.predict(horizon, return_conf_int = True)

        df_preds = get_prediction_dataframe(preds, dates, self.boxcox)

        if plot:

            plot_forecast(df_preds, df_train, last_obs)

        return df_preds



from datetime import date
from typing import Optional, List

from pydantic import BaseModel, ConfigDict

from mosqlient import types


class UserSchema(BaseModel):
    name: str
    username: str


class ImplementationLanguageSchema(BaseModel):
    language: str


class AuthorSchema(BaseModel):
    user: UserSchema
    institution: Optional[str] = None


class TagSchema(BaseModel):
    id: Optional[int]
    name: str
    color: str


class ModelSchema(BaseModel):
    id: Optional[types.ID]
    name: types.Name
    description: types.Description
    author: AuthorSchema
    repository: types.Repository
    implementation_language: ImplementationLanguageSchema
    disease: types.Disease
    categorical: types.Categorical
    spatial: types.Spatial
    temporal: types.Temporal
    ADM_level: types.ADMLevel
    time_resolution: types.TimeResolution


class PredictionDataRowSchema(BaseModel):
    dates: date
    preds: float
    lower: float
    upper: float
    adm_0: str = "BRA"
    adm_1: Optional[str] = None
    adm_2: Optional[int] = None
    adm_3: Optional[int] = None

    class Config:
        json_encoders = {
            date: lambda v: v.strftime('%Y-%m-%d')
        }

    def dict(self, **kwargs):
        _d = super().dict(**kwargs)
        _d['dates'] = _d['dates'].strftime('%Y-%m-%d')
        return _d


class PredictionSchema(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        protected_namespaces=()
    )

    id: Optional[types.ID] = None
    model: ModelSchema
    description: types.Description
    commit: types.Commit
    predict_date: types.Date
    data: List[PredictionDataRowSchema]



from typing import Any
from datetime import date


def parse_params(**kwargs) -> dict[str, Any]:
    from mosqlient.registry.models import Base

    params = {}
    for k, v in kwargs.items():
        if isinstance(v, (bool, int, float, str, date, Base)):
            params[k] = str(v)
        elif v is None:
            continue
        else:
            raise TypeError(f"Unknown type f{type(v)}")

    return params



from datetime import date
from typing_extensions import Annotated
from typing import Union, List, Dict, Optional, Literal

import pandas as pd
from pydantic.functional_validators import AfterValidator

from mosqlient.validations import *  # noqa


APP = Annotated[str, AfterValidator(validate_django_app)]

ID = Annotated[int, AfterValidator(validate_id)]
Name = Annotated[str, AfterValidator(validate_name)]
Description = Annotated[str, AfterValidator(validate_description)]
AuthorName = Annotated[str, AfterValidator(validate_author_name)]
AuthorUserName = Annotated[str, AfterValidator(validate_author_username)]
AuthorInstitution = Annotated[
    Optional[str], AfterValidator(validate_author_institution)
]
Repository = Annotated[str, AfterValidator(validate_repository)]
ImplementationLanguage = Annotated[
    str | dict[Literal['language'], str],
    AfterValidator(validate_implementation_language)
]
Disease = Annotated[str, AfterValidator(validate_disease)]
ADMLevel = Annotated[int, AfterValidator(validate_adm_level)]
Temporal = Annotated[bool, AfterValidator(validate_temporal)]
Spatial = Annotated[bool, AfterValidator(validate_spatial)]
Categorical = Annotated[bool, AfterValidator(validate_categorical)]
TimeResolution = Annotated[str, AfterValidator(validate_time_resolution)]
Tags = Annotated[list, AfterValidator(validate_tags)]  # TODO:

Commit = Annotated[str, AfterValidator(validate_commit)]
Date = Annotated[Union[str, date], AfterValidator(validate_date)]
PredictionData = Annotated[
    Union[pd.DataFrame, List[Dict], str, List],
    AfterValidator(validate_prediction_data)
]

UF = Annotated[str, AfterValidator(validate_uf)]
Geocode = Annotated[int, AfterValidator(validate_geocode)]



__all__ = ["upload_model"]

from typing import Literal

import requests

from mosqlient import Client, types
from mosqlient.errors import ModelPostError
from .models import Model, Author


def upload_model(
    name: str,
    description: str,
    repository: str,
    implementation_language: types.ImplementationLanguage,
    disease: Literal["dengue", "chikungunya", "zika"],
    temporal: bool,
    spatial: bool,
    categorical: bool,
    adm_level: Literal[0, 1, 2, 3],
    time_resolution: Literal["day", "week", "month", "year"],
    api_key: str,
    **kwargs
) -> requests.Response:
    client = Client(x_uid_key=api_key)
    author = Author.get(username=client.username)

    if not author:
        raise ModelPostError("Author not found")

    model = Model(
        client=client,
        author=author[0],
        name=name,
        description=description,
        categorical=categorical,
        temporal=temporal,
        spatial=spatial,
        disease=disease,
        repository=repository,
        implementation_language=implementation_language,
        ADM_level=adm_level,
        time_resolution=time_resolution
    )

    return model.post(**kwargs)



import os

# Constants
DJANGO_APPS = ["registry", "datastore", "vis"]
DISEASES = ["chikungunya", "dengue", "zika"]
ADM_LEVELS = [0, 1, 2, 3]
TIME_RESOLUTIONS = ["day", "week", "month", "year"]
PREDICTION_DATA_COLUMNS = [
    "dates",
    "preds",
    "lower",
    "upper",
    "adm_0",
    "adm_1",
    "adm_2",
    "adm_3",
]

API_URL = os.getenv("MOSQLIENT_API_URL", "https://api.mosqlimate.org/api/")
